@@ -1,5 +1,5 @@
 import mmap
-from header import HeaderManager, Header, leap_year, VLR
+from header import HeaderManager, Header, VLR, EVLR
 import datetime
 from struct import pack, unpack, Struct
 from util import *
@@ -17,6 +17,7 @@ class DataProvider():
         self.fileref = False
         self._mmap = False
         self._pmap = False
+        self._evlrmap = False
         self.manager = manager
         self.mode = manager.mode
     
@@ -28,14 +29,19 @@ class DataProvider():
             raise LaspyException("Error opening file")
 
     def point_map(self):
-        '''Create the numpy point map based on the point format.'''
+        '''Create the numpy point map based on the point format.'''   
         if type(self._mmap) == bool:
-            self.map()
+            self.map() 
         self.pointfmt = np.dtype([("point", zip([x.name for x in self.manager.point_format.specs],
-                                [x.np_fmt for x in self.manager.point_format.specs]))])
- 
-        self._pmap = np.frombuffer(self._mmap, self.pointfmt, 
+                                [x.np_fmt for x in self.manager.point_format.specs]))]) 
+        if not self.manager.header.version in ("1.3", "1.4"): 
+            self._pmap = np.frombuffer(self._mmap, self.pointfmt, 
                         offset = self.manager.header.data_offset)
+        else:  
+            self._pmap = np.frombuffer(self._mmap, self.pointfmt, 
+                        offset = self.manager.header.data_offset,
+                        count = self.manager.header.point_records_count)
+    
     def close(self, flush = True):
         '''Close the data provider and flush changes if _mmap and _pmap exist.''' 
         if flush and self.manager.has_point_records: 
@@ -108,11 +114,12 @@ class DataProvider():
 
 class FileManager():
     '''Superclass of Reader and Writer, provides most of the data manipulation functionality in laspy.''' 
-    def __init__(self,filename, mode, header = False, vlrs = False): 
+    def __init__(self,filename, mode, header = False, vlrs = False, evlrs = False): 
         '''Build the FileManager object. This is done when opening the file
         as well as upon completion of file modification actions like changing the 
         header padding.'''
         self.vlr_formats = Format("VLR")
+        self.evlr_formats = Format("EVLR")
         self.mode = mode
         self.data_provider = DataProvider(filename, self) 
         self.setup_memoizing()
@@ -121,34 +128,42 @@ class FileManager():
 
         self.point_refs = False
         self._current = 0 
-
+        
+        self.padded = False
         if self.mode in ("r", "rw"):
-            self.setup_read_write(vlrs)
+            self.setup_read_write(vlrs, evlrs)
             return
 
         elif self.mode == "w":
-            self.setup_write(header, vlrs)
+            self.setup_write(header, vlrs, evlrs)
             return
         else:
             raise LaspyException("Append Mode Not Supported")
         
-    def setup_read_write(self, vlrs):
+    def setup_read_write(self, vlrs, evlrs):
         self.data_provider.open("r+b")
         self.data_provider.map() 
         self.header_format = Format("h" + self.grab_file_version())
-        self.vlr_formats = Format("VLR")
-        self.get_header()
+        self.get_header(self.grab_file_version())
         self.populate_vlrs()
         self.point_refs = False
         self.has_point_records = True
         self._current = 0
-        self.point_format = Format(self.header.data_format_id) 
+        
+        self.correct_rec_len()
+
         self.data_provider.point_map()
+        if self.header.version in ("1.3", "1.4"):
+            self.populate_evlrs()
+        else:
+            self.evlrs = []
         if vlrs != False:
             self.set_vlrs(vlrs)
+        if evlrs != False:
+            self.set_evlrs(vlrs)
         return
 
-    def setup_write(self,header, vlrs):
+    def setup_write(self,header, vlrs, evlrs):
         if header == False:
             raise LaspyException("Write mode requires a valid header object.")
         ## No file to store data yet.
@@ -157,32 +172,38 @@ class FileManager():
         self.header_format = header.format 
         self._header = header
         self.header = HeaderManager(header = header, reader = self)
-
-        #if self._header.data_format_id != None:
-        #    self.point_format = Format(str(self._header.data_format_id))
-        #else:
-        #    self.point_format = Format("0")
-        self.point_format = Format(self._header.data_format_id)
+        
 
         self.initialize_file_padding(vlrs)
 
         ## We have a file to store data now.
         self.data_provider.remap()
         self.header.flush()
+        
+        self.correct_rec_len()
         if not vlrs in [[], False]:
             self.set_vlrs(vlrs)
+        if not evlrs in [[], False]:
+            self.set_evlrs(evlrs)
+        else:
+            self.evlrs = []
         if self._header.created_year == 0:
-            self.header.date = datetime.datetime.now()
-        
-        if vlrs != False:
-            self.set_vlrs(vlrs) 
+            self.header.date = datetime.datetime.now() 
         self.populate_vlrs()
-        
+        #self.populate_evlrs()
         return
 
+    def correct_rec_len(self):
+        extrabytes = self.header.data_record_length-Format(self.header.data_format_id).rec_len
+        if extrabytes >= 0:
+            self.point_format = Format(self.header.data_format_id,extra_bytes= extrabytes)
+        else:
+            self.point_format = Format(self.header.data_format_id)
+            self.set_header_property("data_record_length", self.point_format.rec_len) 
 
     def initialize_file_padding(self, vlrs):
         filesize = self._header.format.rec_len
+        self._header.header_size = filesize
         if vlrs != False:
             filesize += sum([len(x) for x in vlrs])
         self.vlr_stop = filesize
@@ -233,18 +254,27 @@ class FileManager():
             return
         self.data_provider._mmap.seek(bytes, 0)
         
-    def read_words(self, name):
+    def read_words(self, name, rec_type = "vlr"):
         '''Read a consecutive sequence of packed binary data, return a single 
         element or list.'''
+        if rec_type == "vlr":
+            source = self.vlr_formats
+        elif rec_type == "evlr":
+            source = self.evlr_formats
+        elif rec_type == "header":
+            source = self.header_format
+        else:
+            raise LaspyException("Invalid source: " + str(rec_type))
         try:
-            dim = self.vlr_formats.lookup[name]
+            dim = source.lookup[name]
         except KeyError:
-            raise LaspyException("Dimension " + name + " not found.")
+            raise LaspyException("Dimension " + name + " not found.")       
         return(self._read_words(dim.fmt, dim.num, dim.length))
+       
 
     def _read_words(self, fmt, num, bytes):
         '''Read a consecutive sequence of packed binary data, return a single
-        element or list'''
+        element or list''' 
         outData = []
         for i in xrange(num):
             dat = self.read(bytes)
@@ -268,15 +298,35 @@ class FileManager():
         self.seek(0, rel = True)
         return(str(v1) +"." +  str(v2))
 
-    def get_header(self):
+    def get_header(self, file_version = 1.2):
         '''Return the header object, or create one if absent.'''
         ## Why is this != neccesary?
         try:
             return(self.header)
-        except:
-            self.header = HeaderManager(header = Header(), reader = self)
+        except: 
+            self.header = HeaderManager(header = Header(file_version), reader = self)
             return(self.header)
 
+    def populate_evlrs(self): 
+        '''Catalogue the extended variable length records'''
+        self.evlrs = []
+        if not self.header.version in ("1.3", "1.4"):
+            return
+        
+    
+        if self.header.version == "1.3":  
+            self.seek(self.header.start_wavefm_data_rec, rel = False)
+            num_vlrs = 1
+        elif self.header.version == "1.4":
+            self.seek(self.header.start_first_evlr, rel = False)
+            num_vlrs = self.get_header_property("num_evlrs")
+        for i in xrange(num_vlrs): 
+            new_vlr = EVLR(None, None, None)
+            new_vlr.build_from_reader(self)
+            self.evlrs.append(new_vlr)  
+        return
+
+
     def populate_vlrs(self): 
         '''Catalogue the variable length records'''
         self.vlrs = []
@@ -290,14 +340,17 @@ class FileManager():
                 raise LaspyException("Error, Calculated Header Data "
                     "Overlaps The Point Records!")
         self.vlr_stop = self.data_provider._mmap.tell()
-        self.header.__dict__["vlrs"] = self.vlrs
         return
 
     def get_vlrs(self):
         '''Populate and return list of :obj:`laspy.header.VLR` objects`.'''
         self.populate_vlrs()
         return(self.vlrs)
-    
+   
+    def get_evlrs(self):
+        self.populate_evlrs()
+        return(self.evlrs)
+
     def get_padding(self):
         '''Return the padding between the end of the VLRs and the beginning of
         the point records'''
@@ -305,23 +358,7 @@ class FileManager():
 
     def get_pointrecordscount(self):
         '''calculate the number of point records'''
-        if self.calc_point_recs != False:
-            return(self.calc_point_recs)
-        
-        if self.header.get_version != "1.3":
-            try:
-                return(len(self.data_provider._pmap))
-            except:
-                new_val =  ((self.data_provider._mmap.size()-
-                    self.header.data_offset)/self.header.data_record_length)
-                self.calc_point_recs = new_val
-                return(new_val)
-        else:
-            raise LaspyException("Version 1.3 is currently broken due to waveform data.")
-            new_val = ((self.header.StWavefmDatPktRec-
-                self.header.data_offset)/self.header.data_record_length)       
-            self.calc_point_recs = new_val
-            return(new_val)
+        return(self.get_header_property("point_records_count"))
 
     def set_input_srs(self):
         pass
@@ -425,7 +462,7 @@ class FileManager():
     def _get_raw_datum(self, rec_offs, spec):
         '''return raw bytes associated with non dimension field (VLR/Header)'''
         return(self.data_provider._mmap[(rec_offs + spec.offs):(rec_offs + spec.offs 
-                        + spec.num*spec.length)])
+                        + spec.num*spec.length)]) 
 
     def _get_datum(self, rec_offs, spec):
         '''Return unpacked data assocaited with non dimension field (VLR/Header)'''
@@ -441,10 +478,13 @@ class FileManager():
     def get_raw_header_property(self, name):
         '''Wrapper for grabbing raw header bytes with _get_raw_datum'''
         spec = self.header_format.lookup[name]
+
         return(self._get_raw_datum(0, spec))
     
     def get_header_property(self, name):
         '''Wrapper for grabbing unpacked header data with _get_datum'''
+        spec = self.header_format.lookup[name]
+
         if name in self.header_changes:
             spec = self.header_format.lookup[name]
             new_val = self._get_datum(0, spec)
@@ -481,29 +521,62 @@ class FileManager():
     
     def get_flag_byte(self):
         return(self.get_dimension("flag_byte"))
-    
+   
+    def get_raw_classification_flags(self):
+        return(self.get_dimension("classification_flags"))
+
+    def get_classification_flags(self): 
+        if not self.header.data_format_id in (6,7,8,9,10):
+            return(self.get_classification())
+        rawDim = self.get_raw_classification_flags()
+        return(np.array([self.packed_str(self.binary_str(x)[0:4]) for x in rawDim]))
+
+    def get_classification_byte(self):
+        return(self.get_dimension("classification_byte"))
+
     def get_return_num(self):
         rawDim = self.get_flag_byte()
-        return(np.array([self.packed_str(self.binary_str(x)[0:3]) for x in rawDim])) 
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            return(np.array([self.packed_str(self.binary_str(x)[0:3]) for x in rawDim])) 
+        elif self.header.data_format_id in (6,7,8,9,10):
+            return(np.array([self.packed_str(self.binary_str(x)[0:4]) for x in rawDim]))
 
     def get_num_returns(self):
         rawDim = self.get_flag_byte()
-        return(np.array([self.packed_str(self.binary_str(x)[3:6]) for x in rawDim]))  
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            return(np.array([self.packed_str(self.binary_str(x)[3:6]) for x in rawDim]))  
+        elif self.header.data_format_id in (6,7,8,9,10):
+            return(np.array([self.packed_str(self.binary_str(x)[4:8]) for x in rawDim]))  
+
+    def get_scanner_channel(self):
+        raw_dim = self.get_raw_classification_flags()
+        if not self.header.data_format_id in (6,7,8,9,10):
+            raise LaspyException("Scanner Channel not present for point format: " + str(self.header.data_format_id))
+        return(np.array([self.packed_str(self.binary_str(x)[4:6]) for x in raw_dim]))
 
     def get_scan_dir_flag(self):
-        rawDim = self.get_flag_byte()
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            rawDim = self.get_flag_byte()
+        elif self.header.data_format_id in (6,7,8,9,10):
+            rawDim = self.get_raw_classification_flags()
         return(np.array([self.packed_str(self.binary_str(x)[6]) for x in rawDim])) 
 
-    def get_edge_flight_line(self):
-        rawDim = self.get_flag_byte()
+    def get_edge_flight_line(self): 
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            rawDim = self.get_flag_byte() 
+        elif self.header.data_format_id in (6,7,8,9,10):
+            rawDim = self.get_raw_classification_flags() 
         return(np.array(([self.packed_str(self.binary_str(x)[7]) for x in rawDim])))
     
     def get_raw_classification(self):
         return(self.get_dimension("raw_classification"))
     
     def get_classification(self): 
-        return(np.array([self.packed_str(self.binary_str(x)[0:5]) 
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            return(np.array([self.packed_str(self.binary_str(x)[0:5]) 
                 for x in self.get_raw_classification()])) 
+        elif self.header.data_format_id in (6,7,8,9,10):
+            return(self.get_dimension("classification_byte"))
 
     def get_synthetic(self):
         return(np.array([self.packed_str(self.binary_str(x)[5]) 
@@ -553,55 +626,66 @@ class FileManager():
             return(self.get_dimension("blue"))
         raise LaspyException("Color is not available for point format: " + str(fmt))
 
+    def get_nir(self):
+        fmt = self.header.data_format_id
+        if fmt in (8,10):
+            return(self.get_dimension("nir"))
+        raise LaspyException("NIR is not available for point format: "+str(fmt))
 
-    def get_wave_packet_descp_idx(self):
+    def get_wave_packet_desc_index(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("wave_packet_descp_idx"))
+            return(self.get_dimension("wave_packet_desc_index"))
         raise LaspyException("Wave Packet Description Index Not"
                        + " Available for Pt Fmt: " + str(fmt))
 
-    def get_byte_offset_to_wavefm_data(self):
+    def get_byte_offset_to_waveform_data(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("byte_offset_to_wavefm_data"))
+            return(self.get_dimension("byte_offset_to_waveform_data"))
         raise LaspyException("Byte Offset to Waveform Data Not"
                        + " Available for Pt Fmt: " + str(fmt))
 
-    def get_wavefm_pkt_size(self):
+    def get_waveform_packet_size(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("wavefm_pkt_size"))
+            return(self.get_dimension("waveform_packet_size"))
         raise LaspyException("Wave Packet Description Index Not"
                        + " Available for Pt Fmt: " + str(fmt))
 
-    def get_return_pt_wavefm_loc(self):
+    def get_return_point_waveform_loc(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("return_pt_wavefm_loc"))
+            return(self.get_dimension("return_point_waveform_loc"))
         raise LaspyException("Return Pointt Waveformm Loc Not"
                        + " Available for Pt Fmt: " +str(fmt))
 
     def get_x_t(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("X_t"))
+            return(self.get_dimension("x_t"))
         raise LaspyException("X(t) Not"
                        + " Available for Pt Fmt: " +str(fmt))
 
     def get_y_t(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("Y_t"))
+            return(self.get_dimension("y_t"))
         raise LaspyException("Y(t) Not"
                        + " Available for Pt Fmt: " +str(fmt))
 
     def get_z_t(self):
         fmt = self.header.data_format_id
         if fmt in (4, 5):
-            return(self.get_dimension("Z_t"))
-        raise LaspyException("Z(t) Not"
+            return(self.get_dimension("z_t"))
+        raise LaspyException("z(t) Not"
                        + " Available for Pt Fmt: " +str(fmt))
+    def get_extra_bytes(self):
+        if "extra_bytes" in self.point_format.lookup.keys():
+            return(self.get_dimension("extra_bytes"))
+        else:
+            raise LaspyException("Extra bytes not present in record")
+
 
 class Reader(FileManager):
     def close(self):
@@ -616,7 +700,7 @@ class Writer(FileManager):
 
     def close(self, ignore_header_changes = False):
         '''Flush changes to mmap and close mmap and fileref''' 
-        if not ignore_header_changes:
+        if (not ignore_header_changes) and (self.has_point_records):
             self.header.update_histogram()
             self.header.update_min_max() 
         self.data_provider.close()
@@ -624,6 +708,70 @@ class Writer(FileManager):
     def __del__(self): 
         self.close()
 
+    def set_evlrs(self, value):
+        if value == False or len(value) == 0:
+            return
+        if not all([x.isEVLR for x in value]):
+            raise LaspyException("set_evlrs requers an iterable object " + 
+                        "composed of :obj:`laspy.header.EVLR` objects.")
+        elif self.mode == "w+":
+            raise NotImplementedError
+        elif self.mode in ("rw", "w"): 
+            if self.header.version == "1.3":
+                old_offset = self.header.start_wavefm_data_rec
+            elif self.header.version == "1.4":
+                old_offset = self.header.start_first_evlr
+                self.set_header_property("num_evlrs", len(value))
+            else:
+                raise LaspyException("Invalid File Version for EVLRs: " + str(self.header.version))
+            # Good we know where the EVLRs should go... but what about if we don't have point records yet?
+            # We can't make that decision yet, in case the user wants to subset the data. 
+            if not self.has_point_records:
+                old_offset = self.header.data_offset
+                if self.header.version == "1.3":
+                    self.header.start_wavefm_data_rec = old_offset
+                else:
+                    if len(value) == 1:
+                        self.header.start_first_evlr = old_offset
+                        self.header.start_wavefm_data_rec = old_offset
+                    else:
+                        wf = self.header.start_wavefm_data_rec
+                        fe = self.header.start_first_evlr
+                        new_wvfm = wf - min(wf, fe) + old_offset
+                        new_frst = fe - min(wf, fe) + old_offset
+                        self.header.start_wavefm_data_rec = new_wvfm
+                        self.header.start_first_evlr = new_frst
+                #if old_offset != 0:
+                #    self.pad_file_for_point_recs(self.get_pointrecordscount())
+                #else:
+                #    old_offset = self.header.data_offset
+                #    self.pad_file_for_point_recs(self.get_pointrecordscount())
+
+            self.data_provider.fileref.seek(0, 0)
+            dat_part_1 = self.data_provider.fileref.read(old_offset)
+            # Manually Close:
+            self.data_provider.close(flush=False)
+            self.data_provider.open("w+b")
+            self.data_provider.fileref.write(dat_part_1)
+            total_evlrs = sum([len(x) for x in value])
+            self.data_provider.fileref.write("\x00"*total_evlrs) 
+            self.data_provider.fileref.close()
+            self.data_provider.open("r+b")
+            self.data_provider.map()
+            self.seek(old_offset, rel = False)
+
+            for evlr in value: 
+                self.data_provider._mmap.write(evlr.to_byte_string())
+
+            if self.has_point_records:
+                self.data_provider.point_map()
+            self.populate_evlrs()
+
+        else:
+            raise(LaspyException("set_evlrs requires the file to be opened in a " + 
+                "write mode, and must be performed before point information is provided." + 
+                "Try closing the file and opening it in rw mode. "))
+ 
     def set_vlrs(self, value): 
         if value == False or len(value) == 0:
             return
@@ -677,7 +825,7 @@ class Writer(FileManager):
                 self.data_provider.remap()
                 return
             else:
-                raise NotImplementedError
+                raise LaspyException("Laspy does not yet support assignment of EVLRs for files which already contain point records.")
         elif self.mode == "rw":
             old_offset = self.header.data_offset
             self.set_header_property("data_offset",
@@ -707,16 +855,32 @@ class Writer(FileManager):
         the data given. This is usually a side effect of set_dimension being 
         called for the first time on a file in write mode. '''
         bytes_to_pad = num_recs * self.point_format.rec_len
-        #old_size = self.data_provider.filesize()
-        old_size = self.header.data_offset 
-        self.data_provider._mmap.flush()
-        self.data_provider.fileref.seek(old_size, 0)
-        self.data_provider.fileref.write("\x00" * (bytes_to_pad))
-        self.data_provider.fileref.flush()
-        self.data_provider.remap(flush = False, point_map = True) 
-        # Write Phase complete, enter rw mode?
+        self.header.point_records_count = num_recs
+        if self.evlrs in [False, []]:
+            #old_size = self.data_provider.filesize()
+            old_size = self.header.data_offset     
+            self.data_provider._mmap.flush()
+            self.data_provider.fileref.seek(old_size, 0)
+            self.data_provider.fileref.write("\x00" * (bytes_to_pad))
+            self.data_provider.fileref.flush()
+            self.data_provider.remap(flush = False, point_map = True) 
+            # Write Phase complete, enter rw mode?
+            self.padded = num_recs
+            return
+        else:
+            d1 = self.data_provider._mmap[0:self.header.data_offset]
+            d2 = self.data_provider._mmap[self.header.data_offset:self.data_provider._mmap.size()]
+            self.data_provider.close()
+            self.data_provider.open("w+b")
+            self.data_provider.fileref.write(d1)
+            self.data_provider.fileref.write("\x00"*(bytes_to_pad))
+            self.data_provider.fileref.write(d2)
+            self.data_provider.close()
+            self.data_provider.remap(point_map = True)
+            self.header.start_wavefm_data_rec += bytes_to_pad
+            if self.header.version == "1.4":
+                self.header.start_first_evlr += bytes_to_pad
         
-        return
 
     def set_dimension(self, name,new_dim):
         '''Set a dimension (X,Y,Z etc) to the given value.'''
@@ -949,51 +1113,103 @@ class Writer(FileManager):
     ########
 
     def set_return_num(self, num):
-        '''Set the binary field return_num in the flag_byte'''
-        flag_byte = self.binary_str_arr(self.get_flag_byte())
-        newBits = self.binary_str_arr(num, 3)
-        outByte = self.bitpack((newBits,flag_byte), ((0,3), (3,8)))
-        self.set_dimension("flag_byte", outByte)
+        '''Set the binary field return_num in the flag_byte''' 
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            flag_byte = self.binary_str_arr(self.get_flag_byte())
+            newBits = self.binary_str_arr(num, 3)
+            outByte = self.bitpack((newBits,flag_byte), ((0,3), (3,8)))
+            self.set_dimension("flag_byte", outByte)
+        elif self.header.data_format_id in (6,7,8,9,10):
+            flag_byte = self.binary_str_arr(self.get_flag_byte())
+            newBits = self.binary_str_arr(num, 4)
+            outByte = self.bitpack((newBits,flag_byte), ((0,4), (4,8)))
+            self.set_dimension("flag_byte", outByte)
         return
 
     def set_num_returns(self, num):
         '''Set the binary field num_returns in the flag_byte'''
-        flag_byte = self.binary_str_arr(self.get_flag_byte())
-        newBits = self.binary_str_arr(num, 3)
-        outByte = self.bitpack((flag_byte,newBits,flag_byte), 
-            ((0,3),(0,3), (6,8)))
-        self.set_dimension("flag_byte", outByte)
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            flag_byte = self.binary_str_arr(self.get_flag_byte())
+            newBits = self.binary_str_arr(num, 3)
+            outByte = self.bitpack((flag_byte, newBits,flag_byte), ((0,3),(0,3),(6,8)))
+            self.set_dimension("flag_byte", outByte)
+        elif self.header.data_format_id in (6,7,8,9,10):
+            flag_byte = self.binary_str_arr(self.get_flag_byte())
+            newBits = self.binary_str_arr(num, 4)
+            outByte = self.bitpack((flag_byte, newBits,flag_byte), ((0,4), (0,4), (4,8)))
+            self.set_dimension("flag_byte", outByte)
         return
 
+    def set_scanner_channel(self, value): 
+        if not self.header.data_format_id in (6,7,8,9,10):
+            raise LaspyException("Scanner Channel not present for point format: " + str(self.header.data_format_id))
+        raw_dim = self.get_raw_classification_flags()
+        new_bits = self.binary_str_arr(value, 2)
+        outByte = self.bitpack((raw_dim, new_bits, raw_dim), ((0,4), (4,6), (6,8)))
+        self.set_raw_classification_flags(outByte) 
+
     def set_scan_dir_flag(self, flag): 
         '''Set the binary field scan_dir_flag in the flag_byte'''
-        flag_byte = self.binary_str_arr(self.get_flag_byte())
-        newBits = self.binary_str_arr(flag, 1)
-        outByte = self.bitpack((flag_byte,newBits,flag_byte), 
-            ((0,6),(0,1), (7,8)))
-        self.set_dimension("flag_byte", outByte)
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            flag_byte = self.binary_str_arr(self.get_flag_byte())
+            newBits = self.binary_str_arr(flag, 1)
+            outByte = self.bitpack((flag_byte,newBits,flag_byte), 
+                ((0,6),(0,1), (7,8)))
+            self.set_dimension("flag_byte", outByte)
+        elif self.header.data_format_id in (6,7,8,9,10):
+            flag_byte = self.binary_str_arr(self.get_raw_classification_flags())
+            newBits = self.binary_str_arr(flag, 1)
+            outByte = self.bitpack((flag_byte,newBits,flag_byte), 
+                ((0,6),(0,1), (7,8)))
+            self.set_dimension("classification_flags", outByte)
         return
 
     def set_edge_flight_line(self, line):
         '''Set the binary field edge_flight_line in the flag_byte'''
-        flag_byte = self.binary_str_arr(self.get_flag_byte())
-        newBits = self.binary_str_arr(line, 1)
-        outByte = self.bitpack((flag_byte,newBits), 
-            ((0,7),(0,1)))
-        self.set_dimension("flag_byte", outByte)
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            raw_dim = self.binary_str_arr(self.get_flag_byte())
+            newBits = self.binary_str_arr(line, 1)
+            outByte = self.bitpack((raw_dim, newBits), ((0,7), (0,1)))
+            self.set_dimension("flag_byte", outByte)
+        elif self.header.data_format_id in (6,7,8,9,10):
+            raw_dim = self.binary_str_arr(self.get_raw_classification_flags()) 
+            newBits = self.binary_str_arr(line, 1)
+            outByte = self.bitpack((raw_dim, newBits), ((0,7), (0,1)))
+            self.set_dimension("classification_flags", outByte)
+        return
+
+    def set_classification_byte(self, value):
+        self.set_dimension("classification_byte", value)
+
+    def set_raw_classification_flags(self, value):
+        self.set_dimension("classification_flags",value)
+    
+    def set_classification_flags(self, value): 
+        if not self.header.data_format_id in (6,7,8,9,10):
+            self.set_classification(value)
+            return
+        rawDim = self.binary_str_arr(self.get_raw_classification_flags()) 
+        new_bits = self.binary_str_arr(value, 4) 
+        outbyte = self.bitpack((new_bits, rawDim), ((0,4), (4,8)))
+        self.set_raw_classification_flags(outbyte)
         return
 
     def set_raw_classification(self, classification):
         '''Set the entire classification byte at once. This is faster than setting the binary fields individually, 
         but care must be taken that the values mean what you think they do. '''
         self.set_dimension("raw_classification", classification)
-           
+
     def set_classification(self, classification):
-        '''Set the binary classification field inside the raw classification byte'''
-        class_byte = self.binary_str_arr(self.get_raw_classification())
-        new_bits = self.binary_str_arr(classification, 5)
-        out_byte = self.bitpack((new_bits, class_byte), ((0,5),(5,8)))
-        self.set_dimension("raw_classification", out_byte)
+        '''Point Formats <6: Set the binary classification field inside the raw classification byte
+           Point Formats >5: Set the classification byte.
+        '''
+        if self.header.data_format_id in (0,1,2,3,4,5):
+            class_byte = self.binary_str_arr(self.get_raw_classification())
+            new_bits = self.binary_str_arr(classification, 4)
+            out_byte = self.bitpack((new_bits, class_byte), ((0,5), (5,8)))
+            self.set_raw_classification(out_byte)
+        elif self.header.data_format_id in (6,7,8,9,10):
+            self.set_dimension("classification_byte", classification)
         return
 
     def set_synthetic(self, synthetic):
@@ -1071,65 +1287,78 @@ class Writer(FileManager):
             return
         raise LaspyException("Color Data Not Available for Point Format: " + str(vsn))
 
-    def set_wave_packet_descp_idx(self, idx):
-        '''Wrapper for set_dimension("wave_packet_descp_idx") This is not currently functional, 
+    def set_nir(self, value):
+        fmt = self.header.data_format_id
+        if fmt in (8,10):
+            self.get_dimension("nir", value)
+        raise LaspyException("NIR is not available for point format: "+str(fmt))
+
+    def set_wave_packet_desc_index(self, idx):
+        '''Wrapper for set_dimension("wave_packet_desc_index") This is not currently functional, 
         since addition of waveform data broke the numpy point map.'''
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("wave_packet_descp_index", idx)
+            self.set_dimension("wave_packet_desc_index", idx)
             return
         raise LaspyException("Waveform Packet Description Index Not Available for Point Format: " + str(vsn))
 
-    def set_byte_offset_to_wavefm_data(self, idx):
-        '''Wrapper for set_dimension("byte_offset_to_wavefm_data"), not currently functional,
+    def set_byte_offset_to_waveform_data(self, idx):
+        '''Wrapper for set_dimension("byte_offset_to_waveform_data"), not currently functional,
         because addition of waveform data broke the numpy point map.'''
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("byte_offset_to_wavefm_data", idx)
+            self.set_dimension("byte_offset_to_waveform_data", idx)
             return
         raise LaspyException("Byte Offset To Waveform Data Not Available for Point Format: " + str(vsn))
 
 
     
-    def set_wavefm_pkt_size(self, size):
-        '''Wrapper for set_dimension("wavefm_pkt_size"), not currently functional, because
+    def set_waveform_packet_size(self, size):
+        '''Wrapper for set_dimension("waveform_packet_size"), not currently functional, because
         addition of waveform data broke the numpy point map.'''
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("wavefm_pkt_size", size)
+            self.set_dimension("waveform_packet_size", size)
             return
         raise LaspyException("Waveform Packet Size Not Available for Point Format: " + str(vsn))
     
-    def set_return_pt_wavefm_loc(self, loc):
-        '''Wrapper for set_dimension("return_pt_wavefm_loc"), not currently functional,
+    def set_return_point_waveform_loc(self, loc):
+        '''Wrapper for set_dimension("return_point_waveform_loc"), not currently functional,
         because addition of waveform data broke the numpy point map.'''
         
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("return_pt_wavefm_loc", loc)
+            self.set_dimension("return_point_waveform_loc", loc)
             return
         raise LaspyException("Return Point Waveform Loc Not Available for Point Format: " + str(vsn))
     
     def set_x_t(self, x):
-        '''Wrapper for set_dimension("X_t")'''
+        '''Wrapper for set_dimension("x_t")'''
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("X_t", x)
+            self.set_dimension("x_t", x)
             return
-        raise LaspyException("X_t Not Available for Point Format: " + str(vsn))
+        raise LaspyException("x_t Not Available for Point Format: " + str(vsn))
 
     def set_y_t(self, y):
-        '''Wrapper for set_dimension("Y_t")'''
+        '''Wrapper for set_dimension("y_t")'''
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("Y_t", y)
+            self.set_dimension("y_t", y)
             return
-        raise LaspyException("Y_t Not Available for Point Format: " + str(vsn))
+        raise LaspyException("y_t Not Available for Point Format: " + str(vsn))
     
     def set_z_t(self, z):
-        '''Wrapper for set_dimension("Z_t")'''
+        '''Wrapper for set_dimension("z_t")'''
         vsn = self.header.data_format_id
         if vsn in (4, 5):
-            self.set_dimension("Z_t", z)
+            self.set_dimension("z_t", z)
             return
-        raise LaspyException("Z_t Not Available for Point Format: " + str(vsn))
+        raise LaspyException("z_t Not Available for Point Format: " + str(vsn))
+
+    def set_extra_bytes(self, extra_bytes):
+        '''Wrapper for set_dimension("extra_bytes")'''
+        if "extra_bytes" in self.point_format.lookup.keys():
+            self.set_dimension("extra_bytes", extra_bytes)
+        else:
+            raise LaspyException("Extra bytes not present in point format. Try creating a new file with an extended point record length.")
