@@ -13,7 +13,8 @@ class File(object):
                        vlrs=False,
                        mode='r',
                        in_srs=None,
-                       out_srs=None):
+                       out_srs=None,
+                       evlrs = False):
         '''Instantiate a file object to represent an LAS file.
 
         :arg filename: The filename to open
@@ -53,6 +54,7 @@ class File(object):
         self.filename = os.path.abspath(filename)
         self._header = header
         self._vlrs = vlrs
+        self._evlrs = evlrs
         self._mode = mode.lower()
         self.in_srs = in_srs
         self.out_srs = out_srs
@@ -71,7 +73,7 @@ class File(object):
                 self._reader = base.Reader(self.filename,mode= self._mode)            
                 self._header = self._reader.get_header() 
             else: 
-                raise util.LaspyException("Headers must currently be stored in the file.")
+                raise util.LaspyException("Headers must currently be stored in the file, you provided: " + str(self._header))
                 self._reader = base.Reader(self.filename, mode = self._mode, header=self._header)
 
             if self.in_srs:
@@ -85,20 +87,26 @@ class File(object):
                 self._reader = self._writer
                 self._header = self._reader.get_header()
             else:
-                raise util.LaspyException("Headers must currently be stored in the file.")
+                raise util.LaspyException("Headers must currently be stored in the file, you provided: " + str(self._header))
     
         if self._mode == 'w': 
             if self._header == None:
                 raise util.LaspyException("Creation of a file in write mode requires a header object.")  
             if isinstance(self._header,  header.HeaderManager):
                 vlrs = self._header.vlrs
+                evlrs = self._header.evlrs
                 self._header = self._header.copy() 
                 if self._vlrs != False:
                     self._vlrs.extend(vlrs)
                 else:
                     self._vlrs = vlrs
+                if self._evlrs != False:
+                    self._evlrs.extend(evlrs)
+                else:
+                    self._evlrs = evlrs
             self._writer = base.Writer(self.filename, mode = "w",
-                                      header = self._header,vlrs = self._vlrs)
+                                      header = self._header, 
+                                      vlrs = self._vlrs, evlrs = self._evlrs)
             self._reader = self._writer
         if self._mode == 'w+':
             raise NotImplementedError
@@ -365,6 +373,22 @@ class File(object):
                               set_classification, None, None)
     Classification = classification
 
+    def get_classification_flags(self):
+        return(self._reader.get_classification_flags())
+    def set_classification_flags(self,value):
+        self.assertWriteMode()
+        self._writer.set_classification_flags(value)
+
+    classification_flags = property(get_classification_flags, set_classification_flags, None, None) 
+
+    def get_scanner_channel(self):
+        return(self._writer.get_scanner_channel())
+    def set_scanner_channel(self, value):
+        self.assertWriteMode()
+        self._writer.set_scanner_channel(value)
+
+    scanner_channel = property(get_scanner_channel, set_scanner_channel, None, None)
+
     def get_synthetic(self):
         return(self._reader.get_synthetic())
     def set_synthetic(self, synthetic):
@@ -460,20 +484,28 @@ class File(object):
     Blue = blue
 
     def get_wave_packet_desc_index(self):
-        return(self._reader.get_wave_packet_descp_idx())
+        return(self._reader.get_wave_packet_desc_index())
     def set_wave_packet_desc_index(self,idx):
         self.assertWriteMode()
-        self._writer.set_wave_packet_descp_idx(idx)
+        self._writer.set_wave_packet_desc_index(idx)
         return
+    
+    def get_nir(self):
+        return(self._reader.get_nir())
+    def set_nir(self, value):
+        self.assertWriteMode()
+        self._writer.set_nir(value)
+
+    nir = property(get_nir, set_nir, None, None)
 
     wave_packet_desc_index = property(get_wave_packet_desc_index,
                                       set_wave_packet_desc_index, None, None)
     
     def get_byte_offset_to_waveform_data(self):
-        return(self._reader.get_byte_offset_to_wavefm_data())
+        return(self._reader.get_byte_offset_to_waveform_data())
     def set_byte_offset_to_waveform_data(self, idx):
         self.assertWriteMode()
-        self._writer.set_byte_offset_to_wavefm_data(idx)
+        self._writer.set_byte_offset_to_waveform_data(idx)
         return
 
     byte_offset_to_waveform_data = property(get_byte_offset_to_waveform_data,
@@ -481,25 +513,25 @@ class File(object):
                                             None, None)
     
     def get_waveform_packet_size(self):
-        return(self._reader.get_wavefm_pkt_size())
+        return(self._reader.get_waveform_packet_size())
     def set_waveform_packet_size(self, size):
         self.assertWriteMode()
-        self._writer.set_wavefm_pkt_size(size)
+        self._writer.set_waveform_packet_size(size)
         return
 
     waveform_packet_size = property(get_waveform_packet_size, 
                                     set_waveform_packet_size, 
                                     None, None)
     
-    def get_return_pt_waveform_loc(self):
-        return(self._reader.get_return_pt_wavefm_loc())
-    def set_return_pt_waveform_loc(self, loc):
+    def get_return_point_waveform_loc(self):
+        return(self._reader.get_return_point_waveform_loc())
+    def set_return_point_waveform_loc(self, loc):
         self.assertWriteMode()
-        self._writer.set_return_pt_wavefm_loc(loc)
+        self._writer.set_return_point_waveform_loc(loc)
         return
 
-    return_pt_waveform_loc = property(get_return_pt_waveform_loc, 
-                                      set_return_pt_waveform_loc,
+    return_point_waveform_loc = property(get_return_point_waveform_loc, 
+                                      set_return_point_waveform_loc,
                                       None, None)
     
     def get_x_t(self):
@@ -528,7 +560,18 @@ class File(object):
 
     z_t = property(get_z_t, set_z_t, None, None)
 
+    def get_extra_bytes(self):
+        return(self._reader.get_extra_bytes())
+    def set_extra_bytes(self, new):
+        self.assertWriteMode()
+        self._writer.set_extra_bytes(new)
 
+    doc = '''It is possible to specify a data_record_length longer than the default, 
+            and the extra space is treated by laspy as raw bytes accessable via this extra_bytes property. 
+            This dimension is only assignable for files in write mode which were instantiated with the appropriate
+            data_record_length from the header.'''
+    extra_bytes = property(get_extra_bytes, set_extra_bytes, None, doc)
+        
     def __iter__(self):
         '''Iterator support (read mode only)
 
