@@ -18,6 +18,69 @@ def leap_year(year):
 class LaspyHeaderException(Exception):
     pass
 
+class EVLR():
+    ''' An extended VLR as defined in LAS specification 1.4'''
+    def __init__(self, user_id, record_id, VLR_body, **kwargs):
+        '''Build the EVLR using the required arguments user_id, record_id, and
+        VLR_body. The user can also specify the reserved and description fields if desired.'''
+        self.user_id = str(user_id) + "\x00"*(16-len(str(user_id)))
+        self.record_id = record_id
+        self.VLR_body = VLR_body
+        try: 
+            self.rec_len_after_header = len(self.VLR_body) 
+        except(TypeError):
+            self.rec_len_after_header = 0
+        if "description" in kwargs:
+            self.description = str(kwargs["description"]) + "\x00"*(32-len(kwargs["description"]))
+        else:
+            self.description = "\x00"*32
+        if "reserved" in kwargs:
+            self.reserved = kwargs["reserved"]
+        else:
+            self.reserved = 0 
+        self.fmt = util.Format("EVLR")
+        self.isEVLR = True
+
+    def build_from_reader(self, reader):
+        '''Build an evlr from a reader capable of reading in the data.'''
+        self.reserved = reader.read_words("reserved", "evlr")
+        self.user_id = "".join(reader.read_words("user_id", "evlr"))
+        self.record_id = reader.read_words("record_id", "evlr")
+        self.rec_len_after_header = reader.read_words("rec_len_after_header", "evlr")
+        self.description = "".join(reader.read_words("description", "evlr"))
+        self.VLR_body = reader.read(self.rec_len_after_header)
+        ### LOGICAL CONTENT ###
+        self.isEVLR = True
+        self.fmt = reader.evlr_formats
+
+    def __len__(self):
+        '''Return the size of the evlr object in bytes'''
+        return self.rec_len_after_header + 60
+
+    def pack(self, name, val): 
+        '''Pack an EVLR field into bytes.'''
+        spec = self.fmt.lookup[name]
+        if spec.num == 1:
+            return(pack(spec.full_fmt, val))
+        return(pack(spec.fmt[0]+spec.fmt[1]*len(val), *val))
+    
+    def to_byte_string(self):
+        '''Pack the entire EVLR into a byte string.'''
+        out = (self.pack("reserved", self.reserved) + 
+               self.pack("user_id", self.user_id) + 
+               self.pack("record_id", self.record_id) + 
+               self.pack("rec_len_after_header", self.rec_len_after_header) + 
+               self.pack("description", self.description) +
+               self.VLR_body)
+        diff = (self.rec_len_after_header - len(self.VLR_body))
+        if diff > 0:
+            out += "\x00"*diff
+        elif diff < 0:
+            raise util.LaspyException("Invalid Data in EVLR: too long for specified rec_len." + 
+                                " rec_len_after_header = " + str(self.rec_len_after_header) + 
+                                " actual length = " + str(len(self.VLR_body)))
+        return(out)
+
 
 class VLR():
     '''An object to create/read/store data from LAS Variable Length Records.
@@ -32,17 +95,14 @@ class VLR():
             self.rec_len_after_header = len(self.VLR_body)
         except(TypeError):
             self.rec_len_after_header = 0
-        for key in kwargs:
-            if key == "description":
-                self.description = str(kwargs[key]) + "\x00"*(32-len(kwargs[key]))
-            else:
-                self.description = "\x00"*32
-            if key == "reserved":
-                self.reserved = kwargs[key]
-            else:
-                self.reserved = 0
-            if not key in ["description", "reserved"]:
-                raise util.LaspyException("Unknown Argument to VLR: " + str(key))
+        if "description" in kwargs:
+            self.description = str(kwargs["description"]) + "\x00"*(32-len(kwargs["description"]))
+        else:
+            self.description = "\x00"*32
+        if "reserved" in kwargs:
+            self.reserved = kwargs["reserved"]
+        else:
+            self.reserved = 0 
         self.reserved = 0
         self.isVLR = True
         self.fmt = util.Format("VLR")
@@ -98,24 +158,23 @@ class Header(object):
         kwargs["version_major"] = str(file_version)[0]
         kwargs["version_minor"] = str(file_version)[2]
         self._format = fmt
-        for dim in self.format.specs:
+        for dim in self._format.specs:
             if dim.name in kwargs.keys():
                 self.__dict__[dim.name] = kwargs[dim.name]
             else:
                 self.__dict__[dim.name] = dim.default
 
     def reformat(self, file_version):
-        new_format = Format("h" + str(file_version))
+        new_format = util.Format("h" + str(file_version))
         self.version_major = str(file_version)[0]
-        self.version_minor = str(file_version)[1]
-        
+        self.version_minor = str(file_version)[2]
         # Set defaults for newly available fields.
         for item in new_format.specs:
-            if not item.name in self.__dict__.keys__():
+            if not item.name in self.__dict__.keys():
                 self.__dict__[item.name] = item.default
 
         # Clear no longer available fields. 
-        for item in self.format.specs:
+        for item in self._format.specs:
             if not item.name in new_format.lookup.keys(): 
                 self.__dict__[item.name] = None
         
@@ -242,29 +301,62 @@ class HeaderManager(object):
     gps_time_type = property(get_gps_time_type, set_gps_time_type, None, doc)
 
     def get_waveform_data_packets_internal(self):
-        raise NotImplementedError("LAS Version 1.3 not yet supported.")
+        raw_encoding = self.get_global_encoding()
+        return(self.reader.binary_str(raw_encoding, 16)[1])
+
     def set_waveform_data_packets_internal(self, value):
-        raise NotImplementedError("LAS Version 1.3 not yet supported.")
+        self.assertWriteMode()
+        raw_encoding = self.reader.binary_str(self.get_global_encoding(), 16)
+        self.set_global_encoding(self.reader.packed_str(raw_encoding[0] + str(value) + raw_encoding[2:]))
+        return
+
     waveform_data_packets_internal = property(get_waveform_data_packets_internal, 
                                               set_waveform_data_packets_internal,
                                               None, None)
 
     def get_waveform_data_packets_external(self):
-        raise NotImplementedError("LAS Version 1.3 not yet supported.")
+        raw_encoding = self.get_global_encoding()
+        return(self.reader.binary_str(raw_encoding, 16)[2])
 
     def set_waveform_data_packets_external(self, value):
-        raise NotImplementedError("LAS Version 1.3 not yet supported.")
+        self.assertWriteMode()
+        raw_encoding = self.reader.binary_str(self.get_global_encoding(), 16)
+        self.set_global_encoding(self.reader.packed_str(raw_encoding[0:2], + str(value) + raw_encoding[3:]))
+        return
 
     waveform_data_packets_external = property(get_waveform_data_packets_external,
                                               set_waveform_data_packets_external, 
                                               None, None)
     def get_synthetic_return_num(self):
-        raise NotImplementedError("LAS Version 1.3 not yet supported.")
+        raw_encoding = self.get_global_encoding()
+        return(self.reader.binary_str(raw_encoding, 16)[3])
+
     def set_synthetic_return_num(self, value):
-        raise NotImplementedError("LAS Version 1.3 not yet supported.")
+        self.assertWriteMode()
+        raw_encoding = self.reader.binary_str(self.get_global_encoding(), 16)
+        self.set_global_encoding(self.reader.packed_str(raw_encoding[0:3], + str(value) + raw_encoding[4:]))
+        return
+
     synthetic_return_num = property(get_synthetic_return_num, set_synthetic_return_num, 
                                     None, None)
 
+    def get_wkt(self):
+        if self.data_format_id > 5:
+            raw_encoding = self.get_global_encoding()
+            return(self.reader.binary_str(raw_encoding, 16)[4])
+        else:
+            raise util.LaspyException("WKT not present in data_format_id " + str(self.data_format_id))
+
+    def set_wkt(self, value):
+        self.assertWriteMode()
+        if self.data_format_id > 5:
+            raw_encoding = self.reader.binary_str(self.get_global_encoding(), 16)
+            self.set_global_encoding(self.reader.packed_str(raw_encoding[1:4] + str(value) + raw_encoding[5:16]))
+        else:
+            raise util.LaspyException("WKT not present in data_format_id " + str(self.data_format_id))
+
+    wkt = property(get_wkt, set_wkt, None, None)
+
     def get_projectid(self): 
         p1 = self.reader.get_raw_header_property("proj_id_1")
         p2 = self.reader.get_raw_header_property("proj_id_2")
@@ -508,8 +600,10 @@ class HeaderManager(object):
 
     def get_datarecordlength(self):
         '''Return the size of each point record'''
-        lenDict = {0:20,1:28,2:26,3:34,4:57,5:63}
-        return lenDict[self.data_format_id] 
+        #lenDict = {0:20,1:28,2:26,3:34,4:57,5:63}
+        #return lenDict[self.data_format_id] 
+        return(self.reader.get_header_property("data_record_length"))
+
     doc = '''The length of each point record.'''
     data_record_length = property(get_datarecordlength,
                                   None,
@@ -553,7 +647,7 @@ class HeaderManager(object):
         return self.reader.get_pointrecordscount()
 
     def set_pointrecordscount(self, value):
-        if not self.mode in ("w", "w+"):
+        if not self.file_mode in ("w", "w+"):
             raise LaspyHeaderException("File must be open in write or append mode " + 
                                         "to change the number of point records.")
         self.writer.set_header_property("point_records_count", value)
@@ -610,10 +704,11 @@ class HeaderManager(object):
         #for i in rawdata:
         #    histDict[i] += 1        
         #raw_hist = histDict.values()
-        if self.version == "1.3":
-            raw_hist = np.histogram(rawdata, bins = [1,2,3,4,5,6,7,8])
+        if self.data_format_id in (range(6)):
+            raw_hist = np.histogram(rawdata, bins = range(1,7))
+            
         else:
-            raw_hist = np.histogram(rawdata, bins = [1,2,3,4,5,6])
+            raw_hist = np.histogram(rawdata, bins = range(1,17)) 
         #print("Raw Hist: " + str(raw_hist))
         #t = raw_hist[0][4]
         #for ret in [3,2,1,0]:
@@ -755,6 +850,64 @@ class HeaderManager(object):
     '''
     max = property(get_max, set_max, None, doc)
     
+
+    def get_start_wavefm_data_record(self):
+        if not self.version in ("1.3", "1.4"):
+            raise util.LaspyException("Waveform data not present in version: " + self.version)
+        return(self.reader.get_header_property("start_wavefm_data_rec"))
+
+    def set_start_wavefm_data_record(self, value):
+        self.assertWriteMode()
+        if not self.version in ("1.3", "1.4"):
+            raise util.LaspyException("Waveform data not present in version: " + self.version)
+        self.reader.set_header_property("start_wavefm_data_rec", value)
+
+    start_wavefm_data_rec = property(get_start_wavefm_data_record, set_start_wavefm_data_record, None, None)
+
+    def get_start_first_evlr(self):
+        if not self.version == "1.4":
+            raise util.LaspyException("EVLRs are present explicitly only in version 1.4")
+        return(self.reader.get_header_property("start_first_evlr"))
+
+    def set_start_first_evlr(self, value):
+        if not self.version == "1.4":
+            raise util.LaspyException("EVLRs are present explicitly only in version 1.4")
+        self.reader.set_header_property("start_first_evlr",value)
+        return
+
+    start_first_evlr = property(get_start_first_evlr, set_start_first_evlr, None, None)
+
+    def get_num_EVLRs(self):
+        if not self.version == "1.4":
+            raise util.LaspyException("EVLRs are present explicitly only in version 1.4")
+        return(self.reader.get_header_property("num_EVLRs"))
+
+    def set_num_EVLRs(self, value):
+        if not self.version == "1.4":
+            raise util.LaspyException("EVLRs are present explicitly only in version 1.4")
+        self.assertWriteMode()
+        self.reader.set_header_property("num_EVLRs", value)
+
+    def get_legacy_point_records_count(self):
+        if not self.version == "1.4":
+            raise util.LaspyException("Point records count is only denoted as legacy in version 1.4 files.")
+        return(self.reader.get_header_property("legacy_point_records_count"))
+
+    def set_legacy_point_records_count(self, value):
+        if not self.version == "1.4":
+            raise util.LaspyException("Point records count is only denoted as legacy in version 1.4 files.")
+        self.reader.set_header_property("legacy_point_records_count", value)
+
+    def get_legacy_point_return_count(self):
+        if not self.version == "1.4":
+            raise util.LaspyException("Point return count is only denoted as legacy in version 1.4 files.")
+        return(self.reader.get_header_property("legacy_point_return_count"))
+
+    def set_legacy_point_return_count(self, value):
+        if not self.version == "1.4":
+            raise util.LaspyException("Point return count is only denoted as legacy in version 1.4 files.")
+        self.reader.set_header_property("legacy_point_return_count", value)
+
     def xml(self):
         '''Return an xml repreentation of header data. (not implemented)'''
         raise NotImplementedError
@@ -763,7 +916,6 @@ class HeaderManager(object):
         '''Return an etree representation of header data. (not implemented)'''
         raise NotImplementedError
 
-    ### VLR MANIPULATION NOT IMPLEMENTED
     def add_vlr(self, value):
         return
    
@@ -781,6 +933,14 @@ class HeaderManager(object):
         '''
     vlrs = property(get_vlrs, set_vlrs, None, doc)
 
+    def get_evlrs(self):
+        return(self.reader.get_evlrs())
+    def set_evlrs(self, value):
+        self.assertWriteMode()
+        self.reader.set_evlrs(value)
+
+    evlrs = property(get_evlrs, set_evlrs, None, None)
+
     def get_srs(self):
         raise NotImplementedError 
 
